﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Reflection;
using System.Xml.Linq;
using Phoenix.Xna.Framework.Properties;
using Phoenix.Xna.Framework.Reflection;
using Phoenix.Xna.Framework.Serialization.Markup.Objects;
using Phoenix.Xna.Framework.Xml;

namespace Phoenix.Xna.Framework.Serialization.Markup.Fields
{
    /// <summary>
    /// The markup information for a IList field.
    /// </summary>
    internal class ListMarkupField : CollectionMarkupField
    {
        private List<IMarkupObject> _items = new List<IMarkupObject>();     // The items in the collection.

        /// <summary>
        /// Initializes a new instance of the ListMarkupField class.
        /// </summary>
        public ListMarkupField()
            : base()
        {
        }

        #region IMarkupElement Members

        /// <summary>
        /// Serialize the element to XML.
        /// </summary>
        /// <returns>The element markup serialzed as an XElement.</returns>
        public override XElement Serialize()
        {
            XElement element = base.Serialize();

            foreach (IMarkupObject child in _items)
            {
                XElement item = child.Serialize();
                item.SetAttribute(AttributeIDs.IsCollectionItemAttribute, true);
                element.Add(item);
            }

            return element;
        }

        /// <summary>
        /// Deserialize the element from an XElement.
        /// </summary>
        /// <param name="element">The XElement to deserialize this element from.</param>
        public override void Deserialize(XElement element)
        {
            base.Deserialize(element);

            foreach (XElement node in element.Nodes())
            {
                if (node.ReadAttribute<bool>(AttributeIDs.IsCollectionItemAttribute))
                {
                    IMarkupObject markupObject = Serializer.MarkupSelector.GetMarkupObjectFor(node);
                    _items.Add(markupObject);
                    markupObject.Deserialize(node);
                }
            }
        }

        #endregion

        #region IMarkupField Members

        /// <summary>
        /// Extract the markup information from the FieldInfo for this field.
        /// </summary>
        /// <param name="field">The FieldInfo for this field.</param>
        /// <param name="value">The value of the field.</param>
        public override void ExtractMarkup(FieldInfo field, object value)
        {
            base.ExtractMarkup(field, value);

            if (value != null && field.GetFirstAttributeOfType<SerializableIListAttribute>(false) != null)
            {
                if (!typeof(IList).IsAssignableFrom(value.GetType()))
                {
                    throw new InvalidCastException(string.Format(CultureInfo.InvariantCulture, Resources.InvalidCastException, value.GetType().FullName, typeof(IList).Name));
                }

                IList list = value as IList;
                ShouldPersist = list.Count > 0;

                foreach (object item in list)
                {
                    IMarkupObject markupObject = Serializer.MarkupSelector.GetMarkupObjectFor(item);
                    _items.Add(markupObject);
                }                
            }
        }

        /// <summary>
        /// Sets the value of MarkupField to the instance's field value.
        /// </summary>
        /// <param name="instance">The instance to set the field on.</param>
        public override void SetValue(object instance)
        {
            base.SetValue(instance);
            IList list = (IList)_value;

            foreach (IMarkupObject item in _items)
            {
                if (item.Instance == null)
                {
                    item.ConstructInstance();
                }

                list.Add(item.Instance);
            }
        }

        #endregion
    }
}
